@@ -19,6 +19,7 @@
 */
 
 #include "account.h"
+#include "utils_p.h"
 
 #include <QDateTime>
 
@@ -79,6 +80,21 @@ Account::~Account()
     delete d;
 }
 
+bool Account::operator==(const Account &other) const
+{
+    if (d == other.d) {
+        return true;
+    }
+
+    GAPI_COMPARE(accName)
+    GAPI_COMPARE(accessToken)
+    GAPI_COMPARE(refreshToken)
+    GAPI_COMPARE(expireDateTime)
+    GAPI_COMPARE(scopes)
+
+    return true;
+}
+
 QString Account::accountName() const
 {
     return d->accName;
@@ -180,3 +196,8 @@ QUrl Account::bloggerScopeUrl()
 {
     return QUrl(QStringLiteral("https://www.googleapis.com/auth/blogger"));
 }
+
+QUrl Account::mailScopeUrl()
+{
+    return QUrl(QStringLiteral("https://mail.google.com/"));
+}
